-- ------------------------------------------------------------------------------------------------
-- Monitor classes
-- ------------------------------------------------------------------------------------------------

local function L(text) return text end  -- Localization stub

pol_monitorClasses = {}


function Pol_MonitorFactory(monitorType, ...)
	if not pol_monitorClasses[monitorType] then
		error(L("Monitor type '%s' not found"):format(monitorType or "nil"))
	end
	local monitor = pol_monitorClasses[monitorType](...)
	return monitor
end

function Pol_MonitorTypeRegister(type, class)
	if pol_monitorClasses[type] ~= nil then
		-- FIXME: Give error message somewhere
		return
	end
	pol_monitorClasses[type] = class
end

function Pol_MonitorTypeUnRegister(type)
	if pol_monitorClasses[type] == nil then
		-- FIXME: Give error message somewhere
		return
	end
	pol_monitorClasses[type] = nil
end


-- Base -------------------------------------------------------------------------------------------

MonitorBase = polOO.Class(PolEvents)  -- FIXME: NOT GOOD, FIND SOME OTHER WAY FOR PLUGINS TO GET HOLD OF

function MonitorBase:init(buttonTypeOrProfile)
	PolEvents.init(self)
	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile = {buttonType=buttonTypeOrProfile}
	else
		-- We're being recreated from a stored profile
		self.profile = buttonTypeOrProfile
	end
	self.button = Pol_MonitorButtonFactory(self.profile.buttonType)
end

function MonitorBase:delete()
	self.button:delete()
	self.button = nil
	PolEvents.delete(self)
end

function MonitorBase:GetDescription()
	error("Subclasses must override")
end

function MonitorBase:GetUsage()
	error("Subclasses must override")
end

function MonitorBase:GetProfileTable()
	return self.profile
end

function MonitorBase:GetOptionsTable(order, startOrder, name, desc, customEntries)
	order, startOrder = order or 100, startOrder or 1
	
	local optionsTable = {
		type = "group",
		name = name or "",
		desc = desc or "",
		args = {},
	}
	
	if customEntries then
		for k, v in pairs(customEntries.args) do 
			optionsTable.args[k] = v
		end
	end
	
	return optionsTable
end

function MonitorBase:GetAddOptionsTable()
	return {}
end

function MonitorBase:GetButton()
	return self.button
end

function MonitorBase:SetActive(active)	
	-- Actually fade/unfade
	if self.active ~= active then
		self.button:SetActive(active)
		self.active = active
	end
end

-- MonitorDogTag ----------------------------------------------------------------------------------

local DogTag = LibStub("LibDogTag-3.0") --AceLibrary("LibDogTag-3.0")

local MonitorDogTag = polOO.Class(MonitorBase)

local function updaterTron( me )
	me:_UpdateStatus()	
end

function MonitorDogTag:RegisterStrings( )
	dogtag = self.profile.enabledDogTag
	LibStub("LibDogTag-3.0"):RemoveCallback(dogtag, updaterTron)
	LibStub("LibDogTag-3.0"):AddCallback(dogtag, updaterTron, "Unit", { unit = 'player' }, self)

	dogtag = self.profile.timerDogTag 
	LibStub("LibDogTag-3.0"):RemoveCallback(dogtag, updaterTron)
	LibStub("LibDogTag-3.0"):AddCallback(dogtag, updaterTron, "Unit", { unit = 'player' }, self)
end

function MonitorDogTag:delete()
	LibStub("LibDogTag-3.0"):RemoveCallback(self.profile.enabledDogTag, updaterTron)
	LibStub("LibDogTag-3.0"):RemoveCallback(self.profile.timerDogTag, updaterTron)
	MonitorBase.delete(self)
end

function MonitorDogTag:init(buttonTypeOrProfile, ...)
	MonitorBase.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.enabledDogTag = ""
		self.profile.timerDogTag = ""
	end

	self.profile.enabledDogTag = self.profile.enabledDogTag or ""
	self.profile.timerDogTag = self.profile.timerDogTag or ""

	self:RegisterStrings()
end

function MonitorDogTag:SetActive(active)
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end

	if( self.profile.enabledDogTag and self.profile.enabledDogTag ~= "" ) then
		isActive = DogTag:Evaluate(self.profile.enabledDogTag, "Unit", { unit = self.profile.spellTarget or self.profile.monitorUnit or self.currentUnit })
		if( isActive == nil ) then
			--Dogtags will return "" or nil a lot when it means false
			active = false
		elseif( isActive == true or isActive == false ) then
			active = isActive
		elseif( not( tonumber( isActive ) == nil ) ) then
			--numeric is ok
			active = (isActive > 0 )
		elseif( isActive == "True" ) then
			active = true
		else
			--a string type, probably a polmon err.
			print( "Polmonitor DogTag: "..tostring(isActive) )
		end
	end
	
	MonitorBase.SetActive(self, active)
end

function MonitorDogTag:GetOptionsTable(order, startOrder, name, desc, customEntries)

	local optionsTable = {
		args = {
			enabled = {
				type = "input",
				name = L("Enabled DogTag"),
				desc = L("Specify boolean dogtag expression for when this is enabled"),
				usage = L("e.g. \"[HpPercent]\""),
				multiline = true,
				get = function(info) return self:GetEnabledDogTag() end,
				set = function(info, v) self:SetEnabledDogTag(v) end,
				disabled = false,
				order = startOrder + 0,
			},
			timer = {
				type = "input",
				name = L("Label DogTag"),
				desc = L("Specify dogtag expression for the label on this button"),
				usage = L("e.g. \"[HpPercent]\""),
				multiline = true,
				get = function(info) return self:GetTimerDogTag() end,
				set = function(info, v) self:SetTimerDogTag(v) end,
				disabled = false,
				order = startOrder + 1,
			},
		},
	}

	if customEntries then
		for k, v in pairs(customEntries.args) do 
			optionsTable.args[k] = v
		end
	end
	
	return MonitorBase.GetOptionsTable(self, order, startOrder, name, desc, optionsTable)
end


function MonitorDogTag:GetTimerDogTag()
	--brief note: my original idea was to have a timer dog-tag, but I'm not sure if this is ever useful.
	--Instead I'm going for a simpler numeric dogtag, which may be useful.
	return self.profile.timerDogTag
end

function MonitorDogTag:SetTimerDogTag(dogtag)
	self.button:SetCount("")
	self.button:SetName("")
	self.profile.timerDogTag = dogtag
	self:RegisterStrings()
	self:_UpdateStatus()
end

function MonitorDogTag:GetEnabledDogTag()
	return self.profile.enabledDogTag
end

function MonitorDogTag:_UpdateStatus()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	--print( "Polmonitor DogTag: "..tostring(cooldown) )
	if( self.profile.timerDogTag ~= "" ) then
		cooldown = DogTag:Evaluate(self.profile.timerDogTag, "Unit", { unit = self.profile.spellTarget or self.profile.monitorUnit or self.currentUnit })
		if( cooldown == nil or cooldown == "" or cooldown == false ) then
			self.button:SetCount("")
			self.button:SetName("")
		else
			if( not( tonumber( cooldown ) == nil ) ) then
				self.button:SetCount(tostring(cooldown))
				self.button:SetName("")
			else
				self.button:SetCount("")
				self.button:SetName(cooldown)
			end
		end
	end
end

function MonitorDogTag:SetEnabledDogTag(dogtag)
	self.profile.enabledDogTag = dogtag
	self:RegisterStrings()
	self:_UpdateStatus()
end

-- Helper -----------------------------------------------------------------------------------------

local MonitorWithCooldown = polOO.Class(MonitorDogTag)

function MonitorWithCooldown:init(buttonTypeOrProfile, ...)
	MonitorDogTag.init(self, buttonTypeOrProfile)
	self:SetActive(false)
	self:RegisterEvent("ADDON_LOADED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
end

function MonitorWithCooldown:SetActive(active, forceInCombat)
	-- Invert if needed
	local overruled = (self.profile.combatToggle and not (InCombatLockdown() or forceInCombat))
	active = active and not overruled
	
	if self.profile.inverted then
		active = not active
	end

	-- Fade and/or color as needed
	if self:IsTracking() then
		local isUsable, notEnoughPower, notInRange = self.button:IsUsable()
		self.wasTracking = true

		active = active	and not (not isUsable and self:GetTrackUsable() == "fade")
		active = active and not (notEnoughPower and self:GetTrackPower() == "fade")
		active = active and not (notInRange and self:GetTrackRange() == "fade")
		
		if notEnoughPower and self:GetTrackPower() == "color" then
			self.button:SetColorNotEnoughPower()
		elseif notInRange and self:GetTrackRange() == "color" then
			self.button:SetColorNotInRange()
		elseif not isUsable and self:GetTrackUsable() == "color" then
			self.button:SetColorNotUsable()
		else
			self.button:SetColorNormal()
		end
	elseif self.wasTracking then
		-- Return to normal
		self.wasTracking = false
		self.button:SetColorNormal()
	end
	
	MonitorDogTag.SetActive(self, active)
end

function MonitorWithCooldown:SetCooldownTimer(start, duration, enable)
	self.button:SetCooldownTimer(start, duration, enable and not self.profile.hideCooldownToggle)
end

function MonitorWithCooldown:GetOptionsTable(order, startOrder, name, desc, customEntries)

	local optionsTable = {
		type = "group",
		name = name,
		desc = desc,
		handler = self,
		order = order,
		args = {
			combat = {
				type = "toggle",
				name = L("Combat only"),
				desc = L("If on, this buff is only monitored during combat"),
				get = function(info) return self:GetCombatToggle() end,
				set = function(info, v) self:SetCombatToggle(v) end,
				order = startOrder,
			},
			cooldown = {
				type = "toggle",
				name = L("Hide cooldown"),
				desc = L("If on, cooldown display will be suppressed for this monitor"),
				get = function(info) return self:GetHideCooldownToggle() end,
				set = function(info, v) self:SetHideCooldownToggle(v) end,
				order = startOrder + 1,
			},
			invert = {
				type = "toggle",
				name = L("Invert"),
				desc = L("If on, the monitor will show as active when inactive and vice versa"),
				get = function(info) return self:GetInvertedToggle() end,
				set = function(info, v) self:SetInvertedToggle(v) end,
				order = startOrder + 2,
			},
			trackUsable = {
				type = "select",
				style = "dropdown",
				name = L("Track usable"),
				desc = L("How to display that whatever is monitored is not currently usable"),
				values = {fade="Fade", none="None", color="Color"},
				get = function(info) return self:GetTrackUsable() end,
				set = function(info, v) self:SetTrackUsable(v) end,
				hidden = function() return not self:CanTrackUsable() end,
				order = startOrder + 4,
			},
			trackRange = {
				type = "select",
				style = "dropdown",
				name = L("Track range"),
				desc = L("How to display that you're out of range to use this"),
				values = {fade="Fade", none="None", color="Color"},
				get = function(info) return self:GetTrackRange() end,
				set = function(info, v) self:SetTrackRange(v) end,
				hidden = function() return not self:CanTrackRange() end,
				order = startOrder + 5,
			},
			trackPower = {
				type = "select",
				style = "dropdown",
				name = L("Track power"),
				desc = L("How to display that you don't have enough mana/energy/.. to use this"),
				values = {fade="Fade", none="None", color="Color"},
				get = function(info) return self:GetTrackPower() end,
				set = function(info, v) self:SetTrackPower(v) end,
				hidden = function() return not self:CanTrackPower() end,
				order = startOrder + 6,
			},
			texture = {
				type = "input",
				name = L("Texture"),
				desc = L("Set a custom texture for this monitor"),
				get = function(info) return self:GetTexture() end,
				set = function(info, v) self:SetTexture(v) end,
				order = startOrder + 7,
			},
			label = {
				type = "input",
				name = "Hotkey Label",
				desc = "Set a custom hotkey label for this monitor",
				get = function(info) return self:GetLabel() end,
				set = function(info, v) self:SetLabel(v) end,
				order = startOrder + 8,
			},
			
		},
	}
	
	if customEntries then
		for k, v in pairs(customEntries) do 
			optionsTable.args[k] = v
		end
	end
	
	return MonitorDogTag.GetOptionsTable(self, order, startOrder, name, desc, optionsTable)
end

function MonitorWithCooldown:GetHideCooldownToggle()
	return self.profile.hideCooldownToggle
end

function MonitorWithCooldown:SetHideCooldownToggle(value)
	self.profile.hideCooldownToggle = value
	self:_UpdateStatus()
end

function MonitorWithCooldown:GetCombatToggle()
	return self.profile.combatToggle
end

function MonitorWithCooldown:SetCombatToggle(value)
	self.profile.combatToggle = value
	self:_UpdateStatus()
end

function MonitorWithCooldown:GetInvertedToggle()
	return self.profile.inverted
end

function MonitorWithCooldown:SetInvertedToggle(value)
	local oldInverted = self.profile.inverted
	self.profile.inverted = value
	self:SetActive((oldInverted and not self.active) or (not oldInverted and self.active))
end

function MonitorWithCooldown:GetTrackUsable()
	return self:CanTrackUsable() and self.profile.trackUsable or "none"
end

function MonitorWithCooldown:SetTrackUsable(value)
	self.profile.trackUsable = value
	self:_UpdateStatus()
end

function MonitorWithCooldown:CanTrackUsable()
	return true
end

function MonitorWithCooldown:GetTrackRange()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	return self:CanTrackRange() and self.profile.trackRange or "none"
end

function MonitorWithCooldown:SetTrackRange(value)
	self.profile.trackRange = value
	self:_UpdateStatus()
end

function MonitorWithCooldown:CanTrackRange()
	return self.button:HasRange()
end

function MonitorWithCooldown:GetTrackPower()
	return self:CanTrackPower() and self.profile.trackPower or "none"
end

function MonitorWithCooldown:SetTrackPower(value)
	self.profile.trackPower = value
	self:_UpdateStatus()
end

function MonitorWithCooldown:CanTrackPower()
	return self.button:UsesPower()
end

function MonitorWithCooldown:GetTexture()
	return self.profile.texture or self.button:GetTexture()
end

function MonitorWithCooldown:SetTexture(texture)
	self.profile.texture = (texture ~= "") and texture or nil
	self:_UpdateButtonTexture()
end

function MonitorWithCooldown:GetLabel()
	return self.profile.label
end

function MonitorWithCooldown:SetLabel(label)
	self.profile.label = (label ~= "") and label or nil
	self:_UpdateButtonLabel()
end

function MonitorWithCooldown:ADDON_LOADED()
	self.button:SetActive(self.active)
	self:_UpdateStatus()
end

function MonitorWithCooldown:PLAYER_REGEN_DISABLED()
	self:_UpdateStatus(true)
end

function MonitorWithCooldown:PLAYER_REGEN_ENABLED()
	self:_UpdateStatus()
end

function MonitorWithCooldown:PLAYER_FOCUS_CHANGED()
	self:_UpdateStatus()
end

function MonitorWithCooldown:PLAYER_TARGET_CHANGED()
	self:_UpdateStatus()
end

function MonitorWithCooldown:_UpdateStatus()
	MonitorDogTag._UpdateStatus(self)
end

function MonitorWithCooldown:_UpdateButtonTexture()
	self.button:SetTexture(self.profile.texture)
end

function MonitorWithCooldown:_UpdateButtonLabel()
	self.button:SetHotKey(self.profile.label)
end

function MonitorWithCooldown:IsTracking()
	-- We need to poll if we're tracking any of usable, range and power
	return ("nonenonenone" ~= self:GetTrackUsable()..self:GetTrackRange()..self:GetTrackPower())
end


-- Buff -------------------------------------------------------------------------------------------

local MonitorBuff = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("Buff", MonitorBuff)

function MonitorBuff:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.buffName = table.concat({...}, " ")
	end

	self.profile.timerDogTag = self.profile.timerDogTag or false

	self:_AddEmptyEquivalentsToProfileIfNeeded()
	self.button:SetSpell(self.profile.buffName, self.profile.texture)
	
	self.button:SetUnit(self.profile.monitorUnit or "player")
		
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_AURA")
	self:_UpdateStatus()
end

function MonitorBuff:GetDescription()
	return L("Monitors buffs")
end

function MonitorBuff:GetUsage()
	return L("<buffName>")
end

function MonitorBuff:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set buff"),
				desc = L("Specify which buff to monitor"),
				usage = L("<buff name>"),
				get = function(info) return self:GetBuffName() end,
				set = function(info, v) self:SetBuffName(v) end,
				disabled = false,
				order = startOrder + 0,
			},
			addEquivalent = {
				type = "input",
				name = L("Add equivalent"),
				desc = function(info)
					if self.profile.equivalentNames then
						local desc = L("Add another buff as equivalent of this buff.\n%sCurrent equivalents:|r\n%s")
						local equivalents = {}
						for k, v in pairs(self.profile.equivalentNames) do
							if k ~= self.profile.buffName then
								table.insert(equivalents, "- "..k)
							end
						end
						return desc:format("|cFFEECC11", table.concat(equivalents, "\n"))
					else
						return L("Add another buff as equivalent of this buff")
					end
				end,
				usage = L("<buff name>"),
				get = false,
				set = function(info, v) self:AddEquivalent(v) end,
				order = startOrder + 1,
			},
			clearEquivalents = {
				type = "execute",
				name = L("Clear equivalens"),
				desc = L("Remove all configured equivalents for this buff"),
				func = "ClearEquivalents",
				disabled = function() return not self:HasEquivalents() end,
				order = startOrder + 2,
			},
			unit = {
				type = "select",
				style = "dropdown",
				name = L("Set which unit to check on"),
				desc = L("Specify which unit to monitor for the buff (and re-cast on when clicked)"),
				values = {player="Player (self)", pet="Pet", focus="Focus target", target="target"},
				get = function(info) return self:GetUnit() end,
				set = function(info, v) self:SetUnit(v) end,
				order = startOrder + 3,
			},
			mineonly = {
				type = "toggle",
				name = L("My buffs only"),
				desc = L("If on, other people's buffs will be ignored"),
				get = function(info) return self:GetMineOnlyToggle() end,
				set = function(info, v) self:SetMineOnlyToggle(v) end,
				order = startOrder + 4,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+5, self.profile.buffName, L("Manage this buff"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorBuff:GetBuffName()
	return self.profile.buffName
end

function MonitorBuff:SetBuffName(buffName)
	self.profile.buffName = buffName
	self.button:SetSpell(buffName, self.profile.texture)
	self:ClearEquivalents()
	self:_UpdateStatus()
end

function MonitorBuff:AddEquivalent(buffName)
	self.profile.equivalentNames[buffName] = true
	self:_UpdateStatus()
end

function MonitorBuff:ClearEquivalents()
	for k, dummy in pairs(self.profile.equivalentNames) do
		self.profile.equivalentNames[k] = nil
	end	
	self.profile.equivalentNames[self.profile.buffName] = true
	self:_UpdateStatus()
end

function MonitorBuff:HasEquivalents()
	for k, dummy in pairs(self.profile.equivalentNames) do
		if k ~= self.profile.buffName then
			return true
		end
	end
	return false
end

function MonitorBuff:GetUnit()
	return self.profile.monitorUnit or "player"
end

function MonitorBuff:SetUnit(unit)
	self.profile.monitorUnit = unit
	self:_UpdateStatus()
end

function MonitorBuff:GetMineOnlyToggle()
	return self.profile.mineOnly
end

function MonitorBuff:SetMineOnlyToggle(v)
	self.profile.mineOnly = v
	self:_UpdateStatus()
end

function MonitorBuff:CanTrackRange()
	return MonitorWithCooldown.CanTrackRange(self) and self.profile.monitorUnit == "target"
end

function MonitorBuff:PLAYER_FOCUS_CHANGED()
	self:_UpdateStatus()
end

function MonitorBuff:PLAYER_TARGET_CHANGED()
	self:_UpdateStatus()
end

function MonitorBuff:UNIT_AURA(unit)
	if unit == self:GetUnit() then
		self:_UpdateStatus()
	end
end

function MonitorBuff:OnUpdate()
	self:_UpdateStatus()
end

function MonitorBuff:_UpdateStatus(forceInCombat)
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local inCombat = forceInCombat or InCombatLockdown()
	local duration, expiryTime = 0, 0
	local buffFound = false
	local i = 1
	local bcharges = 0

	repeat
		local buffName, _, buffTexture, charges, buffType, tduration, texpiryTime, isMine = UnitBuff(self:GetUnit(), i)
		Pol_Monitor:SetSpellTexture(buffName, buffTexture)
		i = i + 1

		if self.profile.equivalentNames[buffName] and (isMine == "player" or not self.profile.mineOnly) then
			buffFound = true
			bcharges = charges
			if texpiryTime and texpiryTime > 0 then
				duration, expiryTime = tduration, texpiryTime
				break
			end
		end
	until not buffName
	
	-- Start polling if needed (so we can update the "cooldown" if it's refreshed before it runs out)
	if (buffFound or self:IsTracking()) and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(.2)
	elseif self:IsOnUpdateRunning() and not (buffFound or self:IsTracking()) then
		self:StopOnUpdate()
	end
	
	-- Update button visuals
	self.button:SetCooldownTimer((expiryTime or 0) - (duration or 0), duration, buffFound)
	self:SetActive(not buffFound, forceInCombat)
	if( self.profile.timerDogTag == "" ) then
		if( bcharges > 1 ) then
			self.button:SetCount(bcharges)
		else
			self.button:SetCount("")
		end
		
		if self.profile.label ~= "" then
			self.button:SetHotKey(self.profile.label)
		end
	end
	
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorBuff:_UpdateButtonTexture()
	self.button:SetSpell(self.profile.buffName, self.profile.texture)
end

function MonitorBuff:_AddEmptyEquivalentsToProfileIfNeeded()
	if self.profile.equivalentNames == nil then
		self.profile.equivalentNames = {}
		self.profile.equivalentNames[self.profile.buffName] = true
	end
end

-- TargetDebuff -----------------------------------------------------------------------------------

local MonitorTargetDebuff = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("TargetDebuff", MonitorTargetDebuff)

function MonitorTargetDebuff:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)
	
	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.debuffName = table.concat({...}, " ")
	end
	
	self.profile.target = self.profile.target or "target"
	self:_AddEmptyEquivalentsToProfileIfNeeded()
	self.button:SetSpell(self.profile.debuffName, self.profile.texture)
	self.button:SetUnit(self.profile.target)

	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_AURA")
	self:_UpdateStatus()
end

function MonitorTargetDebuff:GetDescription()
	return L("Monitors debuffs on the specified target")
end

function MonitorTargetDebuff:GetUsage()
	return L("<debuffName>")
end

function MonitorTargetDebuff:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set debuff"),
				desc = L("Specify which debuff to monitor"),
				usage = L("<debuff name>"),
				get = function(info) return self:GetDebuffName() end,
				set = function(info, v) self:SetDebuffName(v) end,
				disabled = false,
				order = startOrder + 0,
			},
			addEquivalent = {
				type = "input",
				name = L("Add equivalent"),
				desc = L("Add another buff as equivalent of this buff"),
				usage = L("<buff name>"),
				get = nil,
				set = function(info, v) self:AddEquivalent(v) end,
				order = startOrder + 1,
			},
			clearEquivalents = {
				type = "execute",
				name = L("Clear equivalens"),
				desc = L("Remove all configured equivalents for this buff"),
				func = "ClearEquivalents",
				disabled = function() return not self:HasEquivalents() end,
				order = startOrder + 2,
			},
			target = {
				type = "select",
				style = "dropdown",
				name = L("Set spell target"),
				desc = L("Specify which target to attempt to cast the spell on"),
				values = {target="Current target", focus="Focus target"},
				get = function(info) return self:GetSpellTarget() end,
				set = function(info, v) self:SetSpellTarget(v) end,
				order = startOrder + 3,
			},
			mineonly = {
				type = "toggle",
				name = L("My debuffs only"),
				desc = L("If on, other people's debuffs will be ignored"),
				get = function(info) return self:GetMineOnlyToggle() end,
				set = function(info, v) self:SetMineOnlyToggle(v) end,
				order = startOrder + 4,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+5, self.profile.debuffName, L("Manage this debuff"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorTargetDebuff:GetDebuffName()
	return self.profile.debuffName
end

function MonitorTargetDebuff:SetDebuffName(debuffName)
	self.profile.debuffName = debuffName
	self.button:SetSpell(debuffName, self.profile.texture)
	self:ClearEquivalents()
	self:_UpdateStatus()
end

function MonitorTargetDebuff:AddEquivalent(debuffName)
	self.profile.equivalentNames[debuffName] = true
	self:_UpdateStatus()
end

function MonitorTargetDebuff:ClearEquivalents()
	for k, dummy in pairs(self.profile.equivalentNames) do
		self.profile.equivalentNames[k] = nil
	end	
	self.profile.equivalentNames[self.profile.debuffName] = true
	self:_UpdateStatus()
end

function MonitorTargetDebuff:HasEquivalents()
	for k, dummy in pairs(self.profile.equivalentNames) do
		if k ~= self.profile.debuffName then
			return true
		end
	end
	return false
end

function MonitorTargetDebuff:GetSpellTarget()
	return self.profile.target
end

function MonitorTargetDebuff:SetSpellTarget(value)
	self.profile.target = value
	self.button:SetUnit(value)
end

function MonitorTargetDebuff:GetMineOnlyToggle()
	return self.profile.mineOnly
end

function MonitorTargetDebuff:SetMineOnlyToggle(v)
	self.profile.mineOnly = v
	self:_UpdateStatus()
end

function MonitorTargetDebuff:CanTrackRange()
	return MonitorWithCooldown.CanTrackRange(self) and self.profile.target == "target"
end

function MonitorTargetDebuff:PLAYER_FOCUS_CHANGED()
	self:_UpdateStatus()
end

function MonitorTargetDebuff:PLAYER_TARGET_CHANGED()
	self:_UpdateStatus()
end

function MonitorTargetDebuff:UNIT_AURA(unit)
	if unit == self.profile.target then
		self:_UpdateStatus()
	end
end

function MonitorTargetDebuff:OnUpdate()
	self:_UpdateStatus()
end

function MonitorTargetDebuff:_UpdateStatus(forceInCombat)
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local inCombat = forceInCombat or InCombatLockdown()
	local duration, expiryTime = 0, 0
	local debuffFound = false
	local i = 1
	local bcharges = 0

	repeat
		local debuffName, _, debuffTexture, charges, debuffType, tduration, texpiryTime, isMine = UnitDebuff(self.profile.target, i)
		Pol_Monitor:SetSpellTexture(debuffName, debuffTexture)
		i = i + 1


		if self.profile.equivalentNames[debuffName] and (isMine == "player" or not self.profile.mineOnly) then
			debuffFound = true
			bcharges = charges
			if texpiryTime and texpiryTime > 0 then
				duration, expiryTime = tduration, texpiryTime
				break
			end
		end
	until not debuffName
	
	-- Start polling if needed (so we can update the "cooldown" if it's refreshed before it runs out)
	if (debuffFound or self:IsTracking()) and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(.2)
	elseif self:IsOnUpdateRunning() and not (debuffFound or self:IsTracking()) then
		self:StopOnUpdate()
	end
	
	-- Update button visuals
	self.button:SetCooldownTimer((expiryTime or 0) - (duration or 0), duration, debuffFound)
	self:SetActive(not (
		debuffFound or (self.profile.combatToggle and not inCombat) or
	   	UnitName(self.profile.target) == nil or UnitIsDeadOrGhost(self.profile.target)))

	if( self.profile.timerDogTag == "" ) then
		if( bcharges > 1 ) then
			self.button:SetCount(bcharges)
		else
			self.button:SetCount("")
		end		
					
		if self.profile.label ~= "" then
			self.button:SetHotKey(self.profile.label)
		end
	end

	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorTargetDebuff:_UpdateButtonTexture()
	self.button:SetSpell(self.profile.debuffName, self.profile.texture)
end

function MonitorTargetDebuff:_AddEmptyEquivalentsToProfileIfNeeded()
	if self.profile.equivalentNames == nil then
		self.profile.equivalentNames = {}
		self.profile.equivalentNames[self.profile.debuffName] = true
	end
end


-- SpellCooldown ----------------------------------------------------------------------------------

local MonitorSpellCooldown = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("SpellCooldown", MonitorSpellCooldown)

function MonitorSpellCooldown:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.spellName = table.concat({...}, " ")
		local _, _, _, _, _, _, _, minRange, maxRange = GetSpellInfo(self.profile.spellName)
		if minRange ~= 0 or maxRange ~= 0 then
			self.profile.spellTarget = "target"
		end
	end

	self.profile.spellTarget = self.profile.spellTarget
	self.button:SetSpell(self.profile.spellName, self.profile.texture)
	self.button:SetUnit(self.profile.spellTarget)
	
	self:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
	self:_UpdateStatus()
end

function MonitorSpellCooldown:GetDescription()
	return L("Monitors the player's spell cooldowns")
end

function MonitorSpellCooldown:GetUsage()
	return L("<spell name>")
end

function MonitorSpellCooldown:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set spell"),
				desc = L("Specify which spell to monitor"),
				usage = L("<spell name>"),
				get = function(info) return self:GetSpellName() end,
				set = function(info, v) self:SetSpellName(v) end,
				disabled = true,
				order = startOrder + 0,
			},
			target = {
				type = "select",
				style = "dropdown",
				name = L("Set spell target"),
				desc = L("Specify which target to attempt to cast the spell on"),
				values = {player="Self cast", target ="Current target", pet="Pet", focus="Focus target", none="None"},
				get = function(info) return self:GetSpellTarget() end,
				set = function(info, v) self:SetSpellTarget(v) end,
				order = startOrder + 1,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+2, self.profile.spellName, L("Manage this spell"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorSpellCooldown:GetSpellName()
	return self.profile.spellName
end

function MonitorSpellCooldown:SetSpellName(spellName)
	self.profile.spellName = spellName
	self.button:SetSpell(spellName, self.profile.texture)
	self:_UpdateStatus()
end

function MonitorSpellCooldown:GetSpellTarget()
	return self.profile.spellTarget or "none"
end

function MonitorSpellCooldown:SetSpellTarget(value)
	self.profile.spellTarget = (value ~= "none") and value or nil
	self.button:SetUnit(value)
end

function MonitorSpellCooldown:CanTrackRange()
	return MonitorWithCooldown.CanTrackRange(self) and self.profile.spellTarget == "target"
end

function MonitorSpellCooldown:ACTIONBAR_UPDATE_COOLDOWN()
	self:_UpdateStatus()
end

function MonitorSpellCooldown:OnUpdate()
	self:_UpdateStatus()
end

function MonitorSpellCooldown:_UpdateStatus()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local start, duration, enable = GetSpellCooldown(self.profile.spellName)
	
	if start ~= 0 and not self.isInCooldown and not Pol_Monitor:IgnoreAsGlobalCooldown(duration) then
		self.isInCooldown = true
	elseif self.isInCooldown and start == 0 then
		self.isInCooldown = false
	end
	
	if (self.isInCooldown or self:IsTracking()) and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(.2)
	elseif not self.isInCooldown and not self:IsTracking() and self:IsOnUpdateRunning() then
		self:StopOnUpdate()
	end
	
	self:SetCooldownTimer(start, duration, enable == 1)
	self:SetActive(not self.isInCooldown)
	
	if self.profile.label ~= "" then
		self.button:SetHotKey(self.profile.label)
	end
	
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorSpellCooldown:_UpdateButtonTexture()
	self.button:SetSpell(self.profile.spellName, self.profile.texture)
end


-- ItemCooldown -----------------------------------------------------------------------------------

local MonitorItemCooldown = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("ItemCooldown", MonitorItemCooldown)

function MonitorItemCooldown:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.itemName = table.concat({...}, " ")
	end

	self.button:SetItem(self.profile.itemName, self.profile.texture)
	
	self:RegisterEvent('ADDON_LOADED')
	self:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
	self:RegisterEvent('UNIT_INVENTORY_CHANGED')
	self:_UpdateStatus()
end

function MonitorItemCooldown:GetDescription()
	return L("Monitors item cooldowns")
end

function MonitorItemCooldown:GetUsage()
	return L("<item name>")
end

function MonitorItemCooldown:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set item"),
				desc = L("Specify which item to monitor"),
				usage = L("<item name>"),
				get = function(info) return self:GetItemName() end,
				set = function(info, v) self:SetItemName(v) end,
				disabled = true,
				order = startOrder + 0,
			},
			equipped = {
				type = "toggle",
				name = L("Equipped only"),
				desc = L("If on, this item is only monitored while equipped"),
				get = function(info) return self:GetEquippedToggle() end,
				set = function(info, v) self:SetEquippedToggle(v) end,
				order = startOrder + 1,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+2, self.profile.itemName, L("Manage this item"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorItemCooldown:GetItemName()
	return self.profile.itemName
end

function MonitorItemCooldown:SetItemName(itemName)
	self.profile.itemName = itemName
	self.button:SetItem(itemName, self.profile.texture)
	self:_UpdateStatus()
end

function MonitorItemCooldown:GetEquippedToggle()
	return self.profile.equippedToggle
end

function MonitorItemCooldown:SetEquippedToggle(value)
	self.profile.equippedToggle = value
	self:_UpdateStatus()
end

function MonitorItemCooldown:CanTrackPower()
	return false
end

function MonitorItemCooldown:ADDON_LOADED()
	self:_UpdateStatus()
end

function MonitorItemCooldown:ACTIONBAR_UPDATE_COOLDOWN()
	self:_UpdateStatus()
end

function MonitorItemCooldown:UNIT_INVENTORY_CHANGED(unitId)
	if unitId == "player" then
		self:_UpdateStatus()
	end
end

function MonitorItemCooldown:OnUpdate()
	self:_UpdateStatus()
end

function MonitorItemCooldown:_UpdateStatus()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end

	local start, duration, enable = GetItemCooldown(self.profile.itemName)
	
	if start ~= 0 and not self.isInCooldown and not Pol_Monitor:IgnoreAsGlobalCooldown(duration) then
		self.isInCooldown = true
	elseif self.isInCooldown and start == 0 then
		self.isInCooldown = false
	end

	if (self.isInCooldown or self:IsTracking()) and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(.2)
	elseif not self.isInCooldown and not self:IsTracking() and self:IsOnUpdateRunning() then
		self:StopOnUpdate()
	end

	self:SetCooldownTimer(start, duration, enable == 1)
	self:SetActive(not self.isInCooldown and not (self.profile.equippedToggle and not IsEquippedItem(self.profile.itemName)))
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorItemCooldown:_UpdateButtonTexture()
	self.button:SetItem(self.profile.itemName, self.profile.texture)
end


-- EquipmentSlotCooldown --------------------------------------------------------------------------

local MonitorEquipmentSlotCooldown = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("EquipmentSlotCooldown", MonitorEquipmentSlotCooldown)

local equipmentSlotTable = {
	HeadSlot = L("Head"),
	NeckSlot = L("Neck"),
	ShoulderSlot = L("Shoulder"),
	BackSlot = L("Back"),
	ChestSlot = L("Chest"),
	ShirtSlot = L("Shirt"),
	TabardSlot = L("Tabard"),
	WristSlot = L("Wrist"),
	HandsSlot = L("Hands"),
	WaistSlot = L("Waist"),
	LegsSlot = L("Legs"),
	FeetSlot = L("Feet"),
	Finger0Slot = L("Ring: Left"),
	Finger1Slot = L("Ring: Right"),
	Trinket0Slot = L("Trinket: Upper"),
	Trinket1Slot = L("Trinket: Lower"),
	MainHandSlot = L("Main hand"),
	SecondaryHandSlot = L("Offhand"),
	RangedSlot = L("Ranged"),
	AmmoSlot = L("Ammo"),
}

function MonitorEquipmentSlotCooldown:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.slotName = table.concat({...}, " ")
	end

	self.button:SetInventorySlot(self.profile.slotName, self.profile.texture)
	self.slotId = GetInventorySlotInfo(self.profile.slotName)
	
	self:RegisterEvent('ADDON_LOADED')
	self:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
	self:RegisterEvent('UNIT_INVENTORY_CHANGED')
	self:_UpdateStatus()
end

function MonitorEquipmentSlotCooldown:GetDescription()
	return L("Monitors equipped item cooldowns")
end

function MonitorEquipmentSlotCooldown:GetUsage()
	return nil -- Because the "usage" field doesn't work with the "select" type option :(
end

function MonitorEquipmentSlotCooldown:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set equipment slot"),
				desc = L("Specify which slot to monitor"),
				usage = L("<equipment slot name>"),
				get = function(info) return self:GetPrettifiedName() end,
				set = function(info, v) self:SetSlotName(v) end,
				disabled = true,
				order = startOrder + 0,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+2, self:GetPrettifiedName(), L("Manage this slot"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorEquipmentSlotCooldown:GetAddOptionsTable()
	local options = {
		type = "select",
		style = "dropdown",
		values = equipmentSlotTable,
	}
	return options
end

function MonitorEquipmentSlotCooldown:GetPrettifiedName()
	return equipmentSlotTable[self.profile.slotName] or "Unknown slot: "..self.profile.slotName
end

function MonitorEquipmentSlotCooldown:GetSlotName()
	return self.profile.slotName
end

function MonitorEquipmentSlotCooldown:SetSlotName(slotName)
	self.profile.slotName = slotName
	self.slotId = GetInventorySlotInfo(slotName)
	self.button:SetInventorySlot(self.profile.slotName, self.profile.texture)
	self:_UpdateStatus()
end

function MonitorEquipmentSlotCooldown:ADDON_LOADED()
	self:_UpdateStatus()
end

function MonitorEquipmentSlotCooldown:ACTIONBAR_UPDATE_COOLDOWN()
	self:_UpdateStatus()
end

function MonitorEquipmentSlotCooldown:UNIT_INVENTORY_CHANGED(unitId)
	if unitId == "player" then
		self.button:SetInventorySlot(self.profile.slotName, self.profile.texture)
		self:_UpdateStatus()
	end
end

function MonitorEquipmentSlotCooldown:OnUpdate()
	self:_UpdateStatus()
end

function MonitorEquipmentSlotCooldown:_UpdateStatus()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local start, duration, enable = GetInventoryItemCooldown("player", self.slotId)

	if start ~= 0 and not self.isInCooldown and not Pol_Monitor:IgnoreAsGlobalCooldown(duration) then
		self.isInCooldown = true
	elseif self.isInCooldown and start == 0 then
		self.isInCooldown = false
	end

	if (self.isInCooldown or self:IsTracking()) and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(.2)
	elseif not self.isInCooldown and not self:IsTracking() and self:IsOnUpdateRunning() then
		self:StopOnUpdate()
	end

	self:SetCooldownTimer(start, duration, enable == 1)
	self:SetActive(not self.isInCooldown)
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorEquipmentSlotCooldown:_UpdateButtonTexture()
	self.button:SetInventorySlot(self.profile.slotName, self.profile.texture)
end


-- WeaponEnchant ----------------------------------------------------------------------------------

local MonitorWeaponEnchant = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("WeaponEnchant", MonitorWeaponEnchant)

local weaponSlotTable = {
	MainHandSlot = L("Main hand"),
	SecondaryHandSlot = L("Offhand"),
}

function MonitorWeaponEnchant:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.slotName = table.concat({...}, " ")
	end
	
	self:_UpdateButton()
	self:RegisterEvent('ADDON_LOADED')
	self:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
	self:RegisterEvent('UNIT_INVENTORY_CHANGED')
end

function MonitorWeaponEnchant:GetDescription()
	return L("Monitors weapon enchants")
end

function MonitorWeaponEnchant:GetUsage()
	return nil -- Because the "usage" field doesn't work with the "select" type option :(
end

function MonitorWeaponEnchant:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set weapon slot"),
				desc = L("Specify which slot to monitor"),
				usage = L("<weapon slot name>"),
				get = function(info) return self:GetPrettifiedName() end,
				set = function(info, v) self:SetSlotName(v) end,
				disabled = true,
				order = startOrder + 0,
			},
			cast = {
				type = "input",
				name = L("Set enchant name"),
				desc = L("Specify which enchant to use when clicked (will attempt to /cast it)"),
				usage = L("<enchant name>"),
				get = function(info) return self:GetEnchantName() end,
				set = function(info, v) self:SetEnchantName(v) end,
				order = startOrder + 1,
			},
			use = {
				type = "toggle",
				name = L("Don't /use"),
				desc = L("For most enchants, like rogue poisons, /use is needed to apply it to the right weapon. "..
						  "However, some enchants automatically apply themselves (like shaman weapon enchants), "..
						  "and if your weapon has its own use action leaving this option off will cause it to be "..
						  "triggered when you try to re-apply the enchant"),
				get = function(info) return self:GetDontUseToggle() end,
				set = function(info, v) self:SetDontUseToggle(v) end,
				order = startOrder + 2,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+3, self:GetPrettifiedName(), L("Manage this slot"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorWeaponEnchant:GetAddOptionsTable()
	local options = {
		type = "select",
		style = "dropdown",
		values = weaponSlotTable,
	}
	return options
end

function MonitorWeaponEnchant:GetPrettifiedName()
	return weaponSlotTable[self.profile.slotName] or "Unknown slot: "..self.profile.slotName
end

function MonitorWeaponEnchant:GetSlotName()
	return self.profile.slotName
end

function MonitorWeaponEnchant:SetSlotName(slotName)
	self.profile.slotName = slotName
	self:_UpdateButton()
	self:_UpdateStatus()
end

function MonitorWeaponEnchant:GetEnchantName()
	return self.profile.enchantName
end

function MonitorWeaponEnchant:SetEnchantName(enchantName)
	self.profile.enchantName = enchantName
	self:_UpdateButton()
	self:_UpdateStatus()
end

function MonitorWeaponEnchant:GetDontUseToggle()
	return self.profile.dontUse
end

function MonitorWeaponEnchant:SetDontUseToggle(v)
	self.profile.dontUse = v
	self:_UpdateButton()
	self:_UpdateStatus()
end

function MonitorWeaponEnchant:CanTrackUsable()
	return false
end

function MonitorWeaponEnchant:CanTrackRange()
	return false
end

function MonitorWeaponEnchant:CanTrackPower()
	return false
end

function MonitorWeaponEnchant:ADDON_LOADED()
	self:_UpdateStatus()
end

function MonitorWeaponEnchant:ACTIONBAR_UPDATE_COOLDOWN()
	self:_UpdateStatus()
end

function MonitorWeaponEnchant:UNIT_INVENTORY_CHANGED(unitId)
	if unitId == "player" then
		self:_UpdateButtonTexture()
		self:_UpdateStatus()
	end
end

function MonitorWeaponEnchant:OnUpdate()
	self:_UpdateStatus()
end

function MonitorWeaponEnchant:_UpdateButton()
	local slotId = GetInventorySlotInfo(self.profile.slotName)
	local macrotext = self.profile.enchantName and "/cast "..self.profile.enchantName or nil
	macrotext = macrotext and macrotext..(self.profile.dontUse and "" or "\n/use "..slotId)
	self.button:SetMacroText(macrotext)
	self:_UpdateButtonTexture()
end

function MonitorWeaponEnchant:_UpdateStatus()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo()
	local enchantFound, duration = false, nil
	
	if self.profile.slotName == "MainHandSlot" then
		enchantFound = hasMainHandEnchant
		duration = (mainHandExpiration or 0)/1000
	elseif self.profile.slotName == "SecondaryHandSlot" then
		enchantFound = hasOffHandEnchant
		duration = (offHandExpiration or 0)/1000
	end
	
	-- Update button visuals
	if enchantFound and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(1)
	elseif self:IsOnUpdateRunning() and not enchantFound then
		self:StopOnUpdate()
	end
	
	self:SetCooldownTimer(GetTime(), duration, enchantFound)
	self:SetActive(not enchantFound)
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorWeaponEnchant:_UpdateButtonTexture()
	local slotId, defaultTexture = GetInventorySlotInfo(self.profile.slotName)
	local texturePath = self.profile.texture or GetInventoryItemTexture("player", slotId) or defaultTexture
	self.button:SetTexture(texturePath)
end


-- MacroCooldown -----------------------------------------------------------------------------------

local MonitorMacroCooldown = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("MacroCooldown", MonitorMacroCooldown)

function MonitorMacroCooldown:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.macroName = table.concat({...}, " ")
	end

	self.button:SetMacro(self.profile.macroName, self.profile.texture)
	self:RegisterEvent('ADDON_LOADED')
	self:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
end

function MonitorMacroCooldown:GetDescription()
	return L("Monitors macro cooldowns")
end

function MonitorMacroCooldown:GetUsage()
	return L("<macro name>")
end

function MonitorMacroCooldown:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set macro"),
				desc = L("Specify which macro to monitor"),
				usage = L("<macro name>"),
				get = function(info) return self:GetMacroName() end,
				set = function(info, v) self:SetMacroName(v) end,
				disabled = true,
				order = startOrder + 0,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+2, self.profile.macroName, L("Manage this macro"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorMacroCooldown:GetMacroName()
	return self.profile.macroName
end

function MonitorMacroCooldown:SetMacroName(macroName)
	self.profile.macroName = macroName
	self.button:SetMacro(macroName, self.profile.texture)
	self:_UpdateStatus()
end

function MonitorMacroCooldown:CanTrackUsable()
	return false
end

function MonitorMacroCooldown:CanTrackRange()
	return false
end

function MonitorMacroCooldown:CanTrackPower()
	return false
end

function MonitorMacroCooldown:ADDON_LOADED()
	self:StartOnUpdate(1)  -- Can't find any way but polling to catch castsequence resets etc :(
	self:_UpdateStatus()
end

function MonitorMacroCooldown:ACTIONBAR_UPDATE_COOLDOWN()
	self:_UpdateStatus()
end

function MonitorMacroCooldown:OnUpdate()
	self:_UpdateStatus()
end

function MonitorMacroCooldown:_UpdateStatus()
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local item, spell = GetMacroItem(self.profile.macroName), GetMacroSpell(self.profile.macroName)
	local start, duration, enable
	
	if item then
		start, duration, enable = GetItemCooldown(item)
	elseif spell then
		start, duration, enable = GetSpellCooldown(spell)
	end

	if start ~= 0 and not self.isInCooldown and not Pol_Monitor:IgnoreAsGlobalCooldown(duration) then
		self.isInCooldown = true
		self:StartOnUpdate(.2)
	elseif self.isInCooldown and start == 0 then
		self.isInCooldown = false
		self:StartOnUpdate(1)
	end

	self:SetCooldownTimer(start, duration, enable == 1)
	self:_UpdateButtonTexture()
	self:SetActive(not self.isInCooldown)
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorMacroCooldown:_UpdateButtonTexture()
	if not self.profile.texture then
		local _, texturePath = GetMacroInfo(self.profile.macroName)
		if texturePath ~= self.texturePath then
			self.button:SetTexture(texturePath)
			self.texturePath = texturePath
		end
	end
end

-- RaidwideBuff -------------------------------------------------------------------------------------------

local MonitorRaidwideBuff = polOO.Class(MonitorWithCooldown)
Pol_MonitorTypeRegister("RaidwideBuff", MonitorRaidwideBuff)

function MonitorRaidwideBuff:init(buttonTypeOrProfile, ...)
	MonitorWithCooldown.init(self, buttonTypeOrProfile)

	if type(buttonTypeOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		self.profile.buffName = table.concat({...}, " ")
	end

	self:_AddEmptyEquivalentsToProfileIfNeeded()

	self.button:SetTexture(Pol_Monitor:GetSpellTexture(self.profile.buffName))
	self.currentUnit = "player"
	self:RegisterEvent("UNIT_AURA")
	self:_UpdateStatus()
end

function MonitorRaidwideBuff:GetDescription()
	return L("Monitors single target buffs")
end

function MonitorRaidwideBuff:GetUsage()
	return L("<buffName>")
end

function MonitorRaidwideBuff:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		local customEntries = {
			set = {
				type = "input",
				name = L("Set buff"),
				desc = L("Specify which buff to monitor"),
				usage = L("<buff name>"),
				get = function(info) return self:GetBuffName() end,
				set = function(info, v) self:SetBuffName(v) end,
				disabled = true,
				order = startOrder + 0,
			},
			addEquivalent = {
				type = "input",
				name = L("Add equivalent"),
				desc = function(info)
					if self.profile.equivalentNames then
						local desc = L("Add another buff as equivalent of this buff.\n%sCurrent equivalents:|r\n%s")
						local equivalents = {}
						for k, v in pairs(self.profile.equivalentNames) do
							if k ~= self.profile.buffName then
								table.insert(equivalents, "- "..k)
							end
						end
						return desc:format("|cFFEECC11", table.concat(equivalents, "\n"))
					else
						return L("Add another buff as equivalent of this buff")
					end
				end,
				usage = L("<buff name>"),
				get = false,
				set = function(info, v) self:AddEquivalent(v) end,
				order = startOrder + 1,
			},
			clearEquivalents = {
				type = "execute",
				name = L("Clear equivalens"),
				desc = L("Remove all configured equivalents for this buff"),
				func = "ClearEquivalents",
				disabled = function() return not self:HasEquivalents() end,
				order = startOrder + 2,
			},
			mineonly = {
				type = "toggle",
				name = L("My buffs only"),
				desc = L("If on, other people's buffs will be ignored"),
				get = function(info) return self:GetMineOnlyToggle() end,
				set = function(info, v) self:SetMineOnlyToggle(v) end,
				order = startOrder + 3,
			},
		}
		self.optionsTable = MonitorWithCooldown.GetOptionsTable(
			self, order, startOrder+5, self.profile.buffName, L("Manage this buff"), customEntries)
	end
	
	return self.optionsTable
end

function MonitorRaidwideBuff:GetBuffName()
	return self.profile.buffName
end

function MonitorRaidwideBuff:SetBuffName(buffName)
	self.profile.buffName = buffName
	self.button:SetSpell(buffName, self.profile.texture)
	self:ClearEquivalents()
	self:_UpdateStatus()
end

function MonitorRaidwideBuff:AddEquivalent(buffName)
	self.profile.equivalentNames[buffName] = true
	self:_UpdateStatus()
end

function MonitorRaidwideBuff:ClearEquivalents()
	for k, dummy in pairs(self.profile.equivalentNames) do
		self.profile.equivalentNames[k] = nil
	end	
	self.profile.equivalentNames[self.profile.buffName] = true
	self:_UpdateStatus()
end

function MonitorRaidwideBuff:HasEquivalents()
	for k, dummy in pairs(self.profile.equivalentNames) do
		if k ~= self.profile.buffName then
			return true
		end
	end
	return false
end

function MonitorRaidwideBuff:GetMineOnlyToggle()
	return self.profile.mineOnly
end

function MonitorRaidwideBuff:SetMineOnlyToggle(v)
	self.profile.mineOnly = v
	self:_UpdateStatus()
end

function MonitorRaidwideBuff:CanTrackRange()
	return false
end

function MonitorRaidwideBuff:UNIT_AURA(unit)
	local i = 1
	
	if (unit ~= "target" and unit ~= "focus") then
		repeat
			local buffName, _, buffTexture, charges, buffType, tduration, texpiryTime, isMine = UnitBuff(unit, i)
			Pol_Monitor:SetSpellTexture(buffName, buffTexture)
			i = i + 1

			if self.profile.equivalentNames[buffName] and (isMine == "player" or not self.profile.mineOnly) then
				if self.currentUnit ~= unit then 
					self:StopOnUpdate()
				end
				self.currentUnit = unit
				break
			end
		until not buffName
	end

	self:_UpdateStatus()
end

function MonitorRaidwideBuff:OnUpdate()
	self:_UpdateStatus()
end

function MonitorRaidwideBuff:_UpdateStatus(forceInCombat, unit)
	if( self.button == nil ) then
		--this is a getout incase we try and update after we deleted.
		return
	end
	local inCombat = forceInCombat or InCombatLockdown()
	local duration, expiryTime = 0, 0
	local buffFound = false
	local i = 1
	
	repeat
		local buffName, _, buffTexture, charges, buffType, tduration, texpiryTime, isMine = UnitBuff(self.currentUnit, i)
		Pol_Monitor:SetSpellTexture(buffName, buffTexture)
		i = i + 1

		if self.profile.equivalentNames[buffName] and (isMine == "player" or not self.profile.mineOnly) then
			buffFound = true
			if texpiryTime and texpiryTime > 0 then
				duration, expiryTime = tduration, texpiryTime
				break
			end
		end
	until not buffName
	
	-- Start polling if needed (so we can update the "cooldown" if it's refreshed before it runs out)
	if (buffFound or self:IsTracking()) and not self:IsOnUpdateRunning() then
		self:StartOnUpdate(.2)
	elseif self:IsOnUpdateRunning() and not (buffFound or self:IsTracking()) then
		self:StopOnUpdate()
	end
	
	-- Update button visuals
	self.button:SetCooldownTimer((expiryTime or 0) - (duration or 0), duration, buffFound)
	self:SetActive(not buffFound, forceInCombat)
	MonitorWithCooldown._UpdateStatus(self)
end

function MonitorRaidwideBuff:_UpdateButtonTexture()
	self.button:SetTexture(self.profile.texture)
end

function MonitorRaidwideBuff:_AddEmptyEquivalentsToProfileIfNeeded()
	if self.profile.equivalentNames == nil then
		self.profile.equivalentNames = {}
		self.profile.equivalentNames[self.profile.buffName] = true
	end
end
